#!/usr/bin/env python3

"""Generate a bug report text file to be included with a bug report."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import sys
import platform
import sysconfig
import logging
import subprocess
import zipfile
import os
import os.path
import argparse


logging.basicConfig(filename='bugreport.log', filemode='w', level=logging.INFO)
logger = logging.getLogger()
logger.addHandler(logging.StreamHandler(stream=sys.stdout))


def main(argv):
    """Run a data checkup of the user's choice."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--dir", help="Directory where Garmindb scripts were run", required=True)
    args = parser.parse_args()

    bug_report_txt = args.dir + os.sep + 'bugreport.txt'

    with open(bug_report_txt, 'w') as report:
        report.write(f"sys.version: {sys.version}\n")
        report.write(f"sys.platform: {sys.platform}\n")
        report.write(f"platform.system(): {platform.system()}\n")
        report.write(f"sysconfig.get_platform(): {sysconfig.get_platform()}\n")
        report.write(f"platform.machine(): {platform.machine()}\n")
        report.write(f"platform.architecture(): {platform.architecture()}\n\n")

        output = subprocess.check_output([sys.executable, '-m', 'pip', 'show', 'garmindb'])
        report.write(output.decode())

        requirements_files = ["requirements.txt", 'Fit/requirements.txt' 'utilities/requirements.txt', 'tcx/requirements.txt']

        for requirements_file in requirements_files:
            requirements_file_path = args.dir + os.sep + requirements_file
            report.write(f'\nrequirements {requirements_file_path}\n\n')
            output = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze', requirements_file_path])
            report.write(output.decode())

    with zipfile.ZipFile(args.dir +  os.sep + 'bugreport.zip', 'w') as zip:
        zip.write(bug_report_txt)
        garmindb_log = args.dir + os.sep + 'garmindb.log'
        if os.path.isfile(garmindb_log):
            zip.write(garmindb_log)


if __name__ == "__main__":
    main(sys.argv[1:])



"""Class that manages Garmin Connect download config."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os
import sys
import platform
import subprocess
import datetime
import tempfile

from idbutils import JsonConfig
from fitfile import Sport

from .statistics import Statistics
from idbutils import DbParams


class ConfigException(Exception):
    """Something unexpected happened while handling the configuration."""


class GarminConnectConfigManager(JsonConfig):
    """Class that manages Garmin Connect downloads."""

    temp_dir = tempfile.mkdtemp()
    homedir = os.path.expanduser('~')

    def __init__(self, config_dir=None):
        """Return a new GarminConnectConfigManager instance."""
        self.enabled_statistics = None
        self.config_dir = config_dir or self.__create_dir_if_needed(self.homedir + os.sep + '.GarminDb')
        config_file = self.config_dir + os.sep + 'GarminConnectConfig.json'
        try:
            super().__init__(config_file)
        except Exception as e:
            print(str(e))
            print(f"Missing or bad config: copy GarminConnectConfig.json.example from {os.path.dirname(os.path.abspath(__file__))} to {config_file} and edit it to "
                  "add your Garmin Connect username and password.")
            sys.exit(-1)

    def get_node_value(self, node, leaf):
        node = self.config.get(node)
        if node is not None:
            return node.get(leaf)

    def get_node_value_default(self, node, leaf, default):
        node = self.config.get(node)
        if node is not None:
            return node.get(leaf, default)
        return default

    @classmethod
    def __create_dir_if_needed(cls, dir):
        if not os.path.exists(dir):
            os.makedirs(dir)
        return dir

    def get_session_file(self):
        """Return the path to the session file."""
        return self.config_dir + os.sep + 'garth_session'

    def get_db_type(self):
        """Return the type (SQLite, MySQL, etc) of database that is configured."""
        return self.get_node_value_default('db', 'type', 'sqlite')

    def get_db_user(self):
        """Return the configured username of the database."""
        return self.get_node_value('db', 'user')

    def get_db_password(self):
        """Return the configured password of the database."""
        return self.get_node_value('db', 'password')

    def get_db_host(self):
        """Return the configured hostname of the database."""
        return self.get_node_value('db', 'host')

    def get_db_dir(self, test_dir=False):
        """Return the configured directory of where the database will be stored."""
        return self.__create_dir_if_needed(self.get_base_dir(test_dir) + os.sep + 'DBs')

    def get_db_params(self, test_db=False):
        """Return the database configuration."""
        db_type = self.get_db_type()
        db_params = {
            'db_type' : db_type
        }
        if db_type == 'sqlite':
            db_params['db_path'] = self.get_db_dir(test_db)
        elif db_type == "mysql":
            db_params['db_type'] = 'mysql'
            db_params['db_username'] = self.get_db_user()
            db_params['db_password'] = self.get_db_password()
            db_params['db_host'] = self.get_db_host()
        return DbParams(**db_params)

    def get_base_dir(self, test_dir=False):
        """Return the configured directory of where the data files will be stored."""
        base = self.get_node_value_default('directories', 'base_dir', 'HealthData')
        if test_dir:
            return self.temp_dir + os.sep + base
        if self.get_node_value_default('directories', 'relative_to_home', True):
            return self.__create_dir_if_needed(self.homedir + os.sep + base)
        return self.__create_dir_if_needed(base)

    def get_backup_dir(self):
        """Return the path to the backup directory."""
        return self.__create_dir_if_needed(self.get_base_dir() + os.sep + 'Backups')

    def __get_fit_files_dir(self, test_dir=False):
        return self.get_base_dir(test_dir) + os.sep + 'FitFiles'

    def get_fit_files_dir(self, test_dir=False):
        """Return the configured directory of where the FIT files will be stored creating it if needed."""
        return self.__create_dir_if_needed(self.__get_fit_files_dir(test_dir))

    def __get_monitoring_base_dir(self):
        return self.get_base_dir() + os.sep + 'FitFiles' + os.sep + 'Monitoring'

    def get_monitoring_base_dir(self):
        """Return the configured directory of where the monitoring files will be stored creating it if needed."""
        return self.__create_dir_if_needed(self.__get_monitoring_base_dir())

    def get_monitoring_dir(self, year, test_dir=False):
        """Return the configured directory of where the monitoring files will be stored creating it if needed."""
        return self.__create_dir_if_needed(self.__get_monitoring_base_dir() + os.sep + str(year))

    def get_activities_dir(self, test_dir=False):
        """Return the configured directory of where the activities files will be stored."""
        return self.__create_dir_if_needed(self.__get_fit_files_dir(test_dir) + os.sep + 'Activities')

    def get_sleep_dir(self):
        """Return the configured directory of where the sleep files will be stored."""
        return self.__create_dir_if_needed(self.get_base_dir() + os.sep + 'Sleep')

    def get_weight_dir(self):
        """Return the configured directory of where the weight files will be stored."""
        return self.__create_dir_if_needed(self.get_base_dir() + os.sep + 'Weight')

    def get_rhr_dir(self):
        """Return the configured directory of where the resting heart rate files will be stored."""
        return self.__create_dir_if_needed(self.get_base_dir() + os.sep + 'RHR')

    def get_fitbit_dir(self):
        """Return the configured directory of where the FitBit will be stored."""
        return self.__create_dir_if_needed(self.get_base_dir() + os.sep + 'FitBitFiles')

    def get_mshealth_dir(self):
        """Return the configured directory of where the Microsoft Health will be stored."""
        return self.__create_dir_if_needed(self.get_base_dir() + os.sep + 'MSHealth')

    def get_plugins_dir(self):
        """Return the configured directory where the plugin files are located."""
        return self.__create_dir_if_needed(self.get_base_dir() + os.sep + 'Plugins')

    def get_metric(self):
        """Return the unit system (metric, statute) that is configured."""
        return self.get_node_value_default('settings', 'metric', False)

    def get_secure_password(self):
        """Return the Garmin Connect password from secure storage. On MacOS that is the KeyChain."""
        system = platform.system()
        if system == 'Darwin':
            # This relies on there being a 'internet password' entry for URL https://sso.garmin.com in the login keychain
            domain = 'sso.garmin.com'
            try:
                password = subprocess.check_output(["security", "find-internet-password", "-s", domain, "-w"])
                if password:
                    return password.rstrip()
            except Exception:
                pass
            raise ConfigException(f'Secure password was specified but no "Internet Password" entry was found in the Login Keychain for https://{domain}')

    def get_password_from_file(self):
        """Read the Garmin Connect password from a file."""
        password_file = self.get_node_value('credentials', 'password_file')
        try:
            with open(os.path.expanduser(password_file), 'r') as f:
                return f.read().strip()
        except Exception as e:
            raise ConfigException(f'Failed to read password from file {password_file}: {str(e)}')

    def get_user(self):
        """Return the Garmin Connect username."""
        return self.get_node_value('credentials', 'user')

    def get_password(self):
        """Return the Garmin Connect password."""
        if self.get_node_value_default('credentials', 'secure_password', False):
            return self.get_secure_password()
        if self.get_node_value_default('credentials', 'password_file', None):
            return self.get_password_from_file()
        return self.get_node_value('credentials', 'password')

    def get_garmin_base_domain(self):
        """Return the Garmin base domain to use for api calls."""
        return self.get_node_value_default('garmin', 'domain', "garmin.com")

    def default_display_activities(cls):
        """Return a list of the default activities to display."""
        return [Sport.strict_from_string(activity) for activity in super().default_display_activities]

    def latest_activity_count(self):
        """Return the number of activities to download when getting the latest."""
        return self.get_node_value('data', 'download_latest_activities')

    def all_activity_count(self):
        """Return the number of activities to download when getting all activities."""
        return self.get_node_value('data', 'download_all_activities')

    def stat_start_date(self, stat_type):
        """Return a tuple containing the start date and the number of days to fetch stats from."""
        start_date = self.get_node_value('data', stat_type + '_start_date')
        end_date = self.get_node_value_default('data', stat_type + '_end_date', datetime.datetime.now().date())
        days = (end_date - start_date).days
        return (start_date, days)

    def device_mount_dir(self):
        """Return the directory where the Garmin USB device is mounted."""
        return self.get_node_value('directories', 'mount_dir')

    def __device_garmin_dir(self):
        return self.device_mount_dir() + os.sep + 'garmin'

    def device_settings_dir(self):
        """Return the full path to the settings file on a mounted device."""
        return self.__device_garmin_dir() + os.sep + 'settings'

    def device_monitoring_dir(self):
        """Return the full path to the monitoring files on a mounted device."""
        return self.__device_garmin_dir() + os.sep + 'monitor'

    def device_sleep_dir(self):
        """Return the full path to the sleep files on a mounted device."""
        return self.__device_garmin_dir() + os.sep + 'sleep'

    def device_activities_dir(self):
        """Return the full path to the activities files on a mounted device."""
        return self.__device_garmin_dir() + os.sep + 'activity'

    def download_days_overlap(self):
        """Return the number of days to overlap previously downloaded data when downloading."""
        return self.get_node_value('data', 'download_days_overlap')

    def course_views(self, type):
        """Return a list of course ids to create views for for the given activitiy type."""
        return self.get_node_value('course_views', type)

    def is_stat_enabled(self, statistic):
        """Return whether a particular statistic is enabled or not."""
        return statistic in self.enabled_stats()

    def enabled_stats(self):
        """Return all enabled statistics as a list of string names."""
        if not self.enabled_statistics:
            json_enabled_stats_dict = self.config.get('enabled_stats', {stat_name: True for stat_name in list(Statistics)})
            self.enabled_statistics = [Statistics.from_string(stat_name) for stat_name, stat_enabled in json_enabled_stats_dict.items() if stat_enabled]
        return self.enabled_statistics

    def display_activities(self):
        """Return a list of activities to display."""
        activities_list = self.get_node_value('activities', 'display')
        if not activities_list:
            activities_list = self.get_node_value_default('settings', 'default_display_activities', [])
        return [Sport.strict_from_string(activity) for activity in activities_list]



#!/usr/bin/env python3

"""Class running a checkup against the DB data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import logging
import argparse

from garmindb import GarminConnectConfigManager
from garmindb import format_version
from garmindb import Checkup


logging.basicConfig(filename='checkup.log', filemode='w', level=logging.INFO)
logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


def main(argv):
    """Run a data checkup of the user's choice."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="print the program's version", action='version', version=format_version(sys.argv[0]))
    parser.add_argument("-t", "--trace", help="Turn on debug tracing", type=int, default=0)
    parser.add_argument("-f", "--config", help="Config file path", type=str, default=None)
    checks_group = parser.add_argument_group('Checks')
    checks_group.add_argument("-b", "--battery", help="Check for low battery levels.", action="store_true", default=False)
    checks_group.add_argument("-c", "--course", help="Show statistics from all workouts for a single course.", type=int, default=None)
    checks_group.add_argument("-g", "--goals", help="Run a checkup on the user\'s goals.", action="store_true", default=False)
    checks_group.add_argument("-a", "--all", help="Run a checkup on all of the the user\'s stats.", action="store_true", default=False)
    args = parser.parse_args()

    checkup = Checkup(GarminConnectConfigManager(args.config), debug=args.trace)
    if args.all or args.battery:
        checkup.battery_status()
    if args.course:
        checkup.activity_course(args.course)
    if args.all or args.goals:
        checkup.goals()


if __name__ == "__main__":
    main(sys.argv[1:])



"""Version information for the application."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


python_required = (3, 10, 0)
dev_python_required = (3, 13, 5)
python_tested = (3, 13, 5)
version_info = (3, 6, 5)
prerelease = False


def version_string():
    """Return a version string for a version tuple."""
    return '.'.join(str(digit) for digit in version_info)



"""Test profile file parsing."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import unittest
import logging

import fitfile

from garmindb import GarminConnectConfigManager, GarminUserSettings, GarminPersonalInformation, GarminSocialProfile
from garmindb.garmindb import GarminDb, Attributes


root_logger = logging.getLogger()
handler = logging.FileHandler('profile_file.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestProfileFile(unittest.TestCase):
    """Class for testing profile JSON file parsing."""

    @classmethod
    def setUpClass(cls):
        cls.gc_config = GarminConnectConfigManager()
        cls.file_path = 'test_files'

    def test_parse_usersettings(self):
        db_params = self.gc_config.get_db_params(test_db=True)
        gus = GarminUserSettings(db_params, self.file_path, debug=2)
        if gus.file_count() > 0:
            gus.process()
        gdb = GarminDb(db_params)
        measurement_system = Attributes.measurements_type(gdb)
        self.assertEqual(measurement_system, fitfile.field_enums.DisplayMeasure.statute,
                         'DisplayMeasure expected %r found %r from %r' % (fitfile.field_enums.DisplayMeasure.statute, measurement_system, gus.file_names))

    def test_parse_personalinfo(self):
        db_params = self.gc_config.get_db_params(test_db=True)
        gpi = GarminPersonalInformation(db_params, self.file_path, debug=2)
        if gpi.file_count() > 0:
            gpi.process()
        gdb = GarminDb(db_params)
        locale = Attributes.get_string(gdb, 'locale')
        self.assertEqual(locale, 'en', 'locale expected %r found %r from %r' % ('en', locale, gpi.file_names))

    def test_parse_socialprofile(self):
        db_params = self.gc_config.get_db_params(test_db=True)
        gsp = GarminSocialProfile(db_params, self.file_path, debug=2)
        if gsp.file_count() > 0:
            gsp.process()
        gdb = GarminDb(db_params)
        id = Attributes.get_string(gdb, 'id')
        self.assertEqual(id, '346745092', 'Id expected %r found %r from %r' % ('346745092', id, gsp.file_names))


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Test config handling."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os
import unittest
import logging
import sys

import fitfile
import idbutils
import tcxfile

sys.path.append('..')
from fitfile import version_string as fit_version_string
from idbutils import version_string as utilities_version_string
from tcxfile import version_string as tcx_version_string


root_logger = logging.getLogger()
handler = logging.FileHandler('copy.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestModuleVersions(unittest.TestCase):
    """Class for testing imported module versions match the source tree."""

    def test_versions(self):
        print(f"fitfile version {fitfile.__version__}")
        self.assertEqual(fitfile.__version__, fit_version_string(), f'Fit version actual {fitfile.__version__} expected {fit_version_string()}')
        print(f"idbutils version {idbutils.__version__}")
        self.assertEqual(idbutils.__version__, utilities_version_string(), f'Utilities version actual {idbutils.__version__} expected {utilities_version_string()}')
        print(f"tcxfile version {tcxfile.__version__}")
        self.assertEqual(tcxfile.__version__, tcx_version_string(), f'Tcx version actual {tcxfile.__version__} expected {tcx_version_string()}')


if __name__ == '__main__':
    unittest.main(verbosity=2)


"""Base class for GarminDb plugins."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging


logger = logging.getLogger(__file__)


class MonitoringFitPluginBase():
    """Base class for GarminDb monitoring FIT file plugins."""

    _type = 'MonitoringFit'

    @classmethod
    def matches_monitoring_file(cls, fit_file):
        """Return if the file matches this plugin."""
        return True

    @classmethod
    def init_monitoring(cls, act_db_class, monitoring_table):
        """Initialize an instance of the plugin as an monitoring FIT file plugin."""
        logger.info("Initializing tables for monitoring plugin %s with monitoring table %s", cls.__name__, monitoring_table)



"""Class that takes a parsed activity FIT file object and imports it into a database."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
import sys

import fitfile

from .garmindb import File, ActivitiesDb, Activities, ActivityRecords, ActivityLaps, ActivitySplits, ActivitiesDevices, StepsActivities, \
    CycleActivities, ClimbingActivities, PaddleActivities
from .fit_file_processor import FitFileProcessor


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class ActivityFitFileProcessor(FitFileProcessor):
    """Class that takes a parsed activity FIT file object and imports it into a database."""

    def write_file(self, fit_file):
        """Given a Fit File object, write all of its messages to the DB."""
        self.activity_fit_file_plugins = [plugin for plugin in self.plugin_manager.get_file_processors('ActivityFit', fit_file).values()]
        if len(self.activity_fit_file_plugins):
            root_logger.info("Loaded %d activity plugins %r for file %s", len(self.activity_fit_file_plugins), self.activity_fit_file_plugins, fit_file)
        # Create the db after setting up the plugins so that plugin tables are handled properly
        self.garmin_act_db = ActivitiesDb(self.db_params, self.debug - 1)
        with self.garmin_db.managed_session() as self.garmin_db_session, self.garmin_act_db.managed_session() as self.garmin_act_db_session:
            self._write_message_types(fit_file, fit_file.message_types)

    def _plugin_dispatch(self, handler_name, *args, **kwargs):
        return super()._plugin_dispatch(self.activity_fit_file_plugins, handler_name, *args, **kwargs)

    def _write_device_info_entry(self, fit_file, message_fields):
        device_serial_number = super()._write_device_info_entry(fit_file, message_fields)
        if device_serial_number:
            activity_id = File.id_from_path(fit_file.filename)
            entry = {'activity_id' : activity_id, 'device_serial_number' : device_serial_number}
            if not ActivitiesDevices.s_exists(self.garmin_act_db_session, entry):
                root_logger.debug("_write_device_info_entry activity_id %s, device serial number %s doesn't exist", activity_id, device_serial_number)
                self.garmin_act_db_session.add(ActivitiesDevices(**entry))

    def _write_lap(self, fit_file, message_type, messages):
        """Write all lap messages to the database."""
        for lap_num, message in enumerate(messages):
            self._write_lap_entry(fit_file, message.fields, lap_num)

    def _write_split(self, fit_file, message_type, messages):
        """Write all split messages to the database."""
        for split_num, message in enumerate(messages):
            self._write_split_entry(fit_file, message.fields, split_num)

    def _write_record(self, fit_file, message_type, messages):
        """Write all record messages to the database."""
        for record_num, message in enumerate(messages):
            self._write_record_entry(fit_file, message.fields, record_num)

    def _write_record_entry(self, fit_file, message_fields, record_num):
        # We don't get record data from multiple sources so we don't need to coellesce data in the DB.
        # It's fastest to just write the new data out if it doesn't currently exist.
        activity_id = File.id_from_path(fit_file.filename)
        plugin_record = self._plugin_dispatch('write_record_entry', self.garmin_act_db_session, fit_file, activity_id, message_fields, record_num)
        if not ActivityRecords.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'record' : record_num}):
            record = {
                'activity_id'                       : activity_id,
                'record'                            : record_num,
                'timestamp'                         : fit_file.utc_datetime_to_local(message_fields.timestamp),
                'position_lat'                      : message_fields.get('position_lat'),
                'position_long'                     : message_fields.get('position_long'),
                'distance'                          : message_fields.get('distance'),
                'cadence'                           : message_fields.get('cadence'),
                'hr'                                : message_fields.get('heart_rate'),
                'rr'                                : message_fields.get('respiration_rate'),
                'altitude'                          : message_fields.get('altitude'),
                'speed'                             : message_fields.get('speed'),
                'temperature'                       : message_fields.get('temperature'),
            }
            record.update(plugin_record)
            root_logger.debug("_write_record_entry activity_id %s, record %s doesn't exist", activity_id, record_num)
            self.garmin_act_db_session.add(ActivityRecords(**record))

    def _write_lap_entry(self, fit_file, message_fields, lap_num):
        # we don't get laps data from multiple sources so we don't need to coellesce data in the DB.
        # It's fastest to just write new data out if the it doesn't currently exist.
        activity_id = File.id_from_path(fit_file.filename)
        plugin_lap = self._plugin_dispatch('write_lap_entry', self.garmin_act_db_session, fit_file, activity_id, message_fields, lap_num)
        if not ActivityLaps.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'lap' : lap_num}):
            lap = {
                'activity_id'                       : File.id_from_path(fit_file.filename),
                'lap'                               : lap_num,
                'start_time'                        : fit_file.utc_datetime_to_local(message_fields.start_time),
                'stop_time'                         : fit_file.utc_datetime_to_local(message_fields.timestamp),
                'elapsed_time'                      : message_fields.get('total_elapsed_time'),
                'moving_time'                       : message_fields.get('total_timer_time'),
                'start_lat'                         : message_fields.get('start_position_lat'),
                'start_long'                        : message_fields.get('start_position_long'),
                'stop_lat'                          : message_fields.get('end_position_lat'),
                'stop_long'                         : message_fields.get('end_position_long'),
                'distance'                          : message_fields.get('total_distance'),
                'cycles'                            : message_fields.get('total_cycles'),
                'avg_hr'                            : message_fields.get('avg_heart_rate'),
                'max_hr'                            : message_fields.get('max_heart_rate'),
                'avg_rr'                            : message_fields.get('avg_respiration_rate'),
                'max_rr'                            : message_fields.get('max_respiration_rate'),
                'calories'                          : message_fields.get('total_calories'),
                'avg_cadence'                       : message_fields.get('avg_cadence'),
                'max_cadence'                       : message_fields.get('max_cadence'),
                'avg_speed'                         : message_fields.get('avg_speed'),
                'max_speed'                         : message_fields.get('max_speed'),
                'ascent'                            : message_fields.get('total_ascent'),
                'descent'                           : message_fields.get('total_descent'),
                'max_temperature'                   : message_fields.get('max_temperature'),
                'avg_temperature'                   : message_fields.get('avg_temperature'),
            }
            lap.update(plugin_lap)
            root_logger.debug("writing lap %r for %s", lap, fit_file.filename)
            self.garmin_act_db_session.add(ActivityLaps(**lap))

    def _write_split_entry(self, fit_file, message_fields, split_num):
        # we don't get splits data from multiple sources so we don't need to coellesce data in the DB.
        # It's fastest to just write new data out if the it doesn't currently exist.
        activity_id = File.id_from_path(fit_file.filename)
        plugin_split = self._plugin_dispatch('write_split_entry', self.garmin_act_db_session, fit_file, activity_id, message_fields, split_num)

        if not ActivitySplits.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'split' : split_num}):
            split = {
                'activity_id'                       : File.id_from_path(fit_file.filename),
                'split'                             : split_num,
                'start_time'                        : fit_file.utc_datetime_to_local(message_fields.start_time),
                'stop_time'                         : fit_file.utc_datetime_to_local(message_fields.timestamp),
                'elapsed_time'                      : message_fields.get('total_elapsed_time'),
                'moving_time'                       : message_fields.get('total_timer_time'),
                'avg_hr'                            : message_fields.get('avg_heart_rate'),
                'max_hr'                            : message_fields.get('max_heart_rate'),
                'calories'                          : message_fields.get('total_calories'),
                'ascent'                            : message_fields.get('total_ascent'),
                'descent'                           : message_fields.get('total_descent'),
                'avg_speed'                         : message_fields.get('avg_vertical_speed'),
            }
            bouldering_font_grade = {
                0: '1',
                1: '2',
                2: '3',
                3: '4',
                4: '4+',
                5: '5',
                6: '5+',
                7: '6A',
                8: '6A+',
                9: '6B',
                10: '6B+',
                11: '6C',
                12: '6C+',
                13: '7A',
                14: '7A+',
                15: '7B',
                16: '7B+',
                17: '7C',
                18: '7C+',
                19: '8A',
                20: '8A+',
                21: '8B',
                22: '8B+',
                23: '8C',
                24: '8C+',
                25: '9A',
            }
            indoor_font_grade = {
                0: '1',
                1: '2',
                2: '3',
                3: '4a',
                4: '4b',
                5: '4c',
                6: '5a',
                7: '5b',
                8: '5c',
                9: '6a',
                10: '6a+',
                11: '6b',
                12: '6b+',
                13: '6c',
                14: '6c+',
                15: '7a',
                16: '7a+',
                17: '7b',
                18: '7b+',
                19: '7c',
                20: '7c+',
                21: '8a',
                22: '8a+',
                23: '8b',
                24: '8b+',
                25: '8c',
                26: '8c+',
                27: '9a',
                28: '9a+',
                29: '9b',
            }

            route = {
                'activity_id'   : activity_id,
                'grade'         : message_fields.get('grade'),
                'completed'     : message_fields.get('completed'),
                'falls'         : message_fields.get('falls'),
            }

            if 'grade' in route and route.get('grade') is not None:
                if fitfile.SubSport.bouldering == fit_file.sub_sport_type:
                    route['grade'] = bouldering_font_grade[int(route['grade'])]
                elif fitfile.SubSport.indoor_climbing == fit_file.sub_sport_type:
                    route['grade'] = indoor_font_grade[int(route['grade'])]

            split.update(plugin_split)
            split.update(route)

            # there are some empty splits we want to filter out
            if route.get('grade') is not None and route.get('completed') is not None:
                root_logger.debug("writing split %r for %s", split, fit_file.filename)
                self.garmin_act_db_session.add(ActivitySplits(**split))

    def _write_steps_entry(self, fit_file, activity_id, sub_sport, message_fields):
        steps = {
            'activity_id'                       : activity_id,
            'steps'                             : message_fields.get('total_steps'),
            'avg_pace'                          : fitfile.conversions.perhour_speed_to_pace(message_fields.avg_speed),
            'max_pace'                          : fitfile.conversions.perhour_speed_to_pace(message_fields.max_speed),
            'avg_steps_per_min'                 : message_fields.get('avg_steps_per_min'),
            'max_steps_per_min'                 : message_fields.get('max_steps_per_min'),
            'avg_step_length'                   : message_fields.get('avg_step_length'),
            'avg_vertical_ratio'                : message_fields.get('avg_vertical_ratio'),
            'avg_vertical_oscillation'          : message_fields.get('avg_vertical_oscillation'),
            'avg_gct_balance'                   : message_fields.get('avg_stance_time_balance'),
            'avg_ground_contact_time'           : message_fields.get('avg_stance_time'),
            'avg_stance_time_percent'           : message_fields.get('avg_stance_time_percent'),
        }
        steps.update(self._plugin_dispatch('write_steps_entry', self.garmin_act_db_session, fit_file, activity_id, sub_sport, message_fields))
        root_logger.debug("_write_steps_entry: %r", steps)
        StepsActivities.s_insert_or_update(self.garmin_act_db_session, steps, ignore_none=True, ignore_zero=True)

    def _write_running_entry(self, fit_file, activity_id, sub_sport, message_fields):
        return self._write_steps_entry(fit_file, activity_id, sub_sport, message_fields)

    def _write_walking_entry(self, fit_file, activity_id, sub_sport, message_fields):
        return self._write_steps_entry(fit_file, activity_id, sub_sport, message_fields)

    def _write_hiking_entry(self, fit_file, activity_id, sub_sport, message_fields):
        return self._write_steps_entry(fit_file, activity_id, sub_sport, message_fields)

    def _write_cycling_entry(self, fit_file, activity_id, sub_sport, message_fields):
        ride = {
            'activity_id'   : activity_id,
            'strokes'       : message_fields.get('total_strokes'),
        }
        ride.update(self._plugin_dispatch('write_cycle_entry', self.garmin_act_db_session, fit_file, activity_id, sub_sport, message_fields))
        CycleActivities.s_insert_or_update(self.garmin_act_db_session, ride, ignore_none=True, ignore_zero=True)

    def _write_rock_climbing_entry(self, fit_file, activity_id, sub_sport, message_fields):
        entry = {
            'activity_id'   : activity_id,
            'total_routes': len([s for s in fit_file.split if s.fields.get('grade')]),
        }
        entry.update(self._plugin_dispatch('write_rock_climbing_entry', self.garmin_act_db_session, fit_file, activity_id, sub_sport, message_fields))
        ClimbingActivities.s_insert_or_update(self.garmin_act_db_session, entry, ignore_none=True, ignore_zero=True)

    def _write_stand_up_paddleboarding_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("sup sport entry: %r", message_fields)
        paddle = {
            'activity_id'           : activity_id,
            'strokes'               : message_fields.get('total_strokes'),
            'avg_stroke_distance'   : message_fields.get('avg_stroke_distance'),
        }
        paddle.update(self._plugin_dispatch('write_paddle_entry', self.garmin_act_db_session, fit_file, activity_id, sub_sport, message_fields))
        PaddleActivities.s_insert_or_update(self.garmin_act_db_session, paddle, ignore_none=True, ignore_zero=True)

    def _write_rowing_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("row sport entry: %r", message_fields)
        return self._write_stand_up_paddleboarding_entry(fit_file, activity_id, sub_sport, message_fields)

    def _write_boating_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("boating sport entry: %r", message_fields)

    def _write_fitness_equipment_entry(self, fit_file, activity_id, sub_sport, message_fields):
        try:
            function = getattr(self, '_write_' + sub_sport.name + '_entry')
            function(fit_file, activity_id, sub_sport, message_fields)
        except AttributeError:
            root_logger.info("No sub sport handler type %s from %s: %s", sub_sport, fit_file.filename, message_fields)

    def _write_alpine_skiing_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Skiing sport entry: %r", message_fields)

    def _write_swimming_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Swimming sport entry: %r", message_fields)

    def _write_training_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Training sport entry: %r", message_fields)

    def _write_transition_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Transition sport entry: %r", message_fields)

    def _write_generic_entry(self, fit_file, activity_id, sub_sport, message_fields):
        root_logger.debug("Generic sport entry: %r", message_fields)

    def __choose_sport(self, current_sport, current_sub_sport, new_sport, new_sub_sport):
        sport = fitfile.Sport.strict_from_string(current_sport)
        sub_sport = fitfile.SubSport.strict_from_string(current_sub_sport)
        if new_sport is not None and (sport is None or (not sport.preferred() and new_sport.preferred())):
            sport = new_sport
        if new_sub_sport is not None and (sub_sport is None or (not sub_sport.preferred() and new_sub_sport.preferred())):
            sub_sport = new_sub_sport
        return {'sport' : fitfile.field_enums.name_for_enum(sport), 'sub_sport' : fitfile.field_enums.name_for_enum(sub_sport)}

    def _write_session_entry(self, fit_file, message_fields):
        activity_id = File.id_from_path(fit_file.filename)
        sport = message_fields.sport
        sub_sport = message_fields.sub_sport
        activity = {
            'activity_id'                       : activity_id,
            'start_time'                        : fit_file.utc_datetime_to_local(message_fields.start_time),
            'stop_time'                         : fit_file.utc_datetime_to_local(message_fields.timestamp),
            'elapsed_time'                      : message_fields.total_elapsed_time,
            'moving_time'                       : message_fields.get('total_timer_time'),
            'start_lat'                         : message_fields.get('start_position_lat'),
            'start_long'                        : message_fields.get('start_position_long'),
            'stop_lat'                          : message_fields.get('end_position_lat'),
            'stop_long'                         : message_fields.get('end_position_long'),
            'distance'                          : message_fields.get('total_distance'),
            'cycles'                            : message_fields.get('total_cycles'),
            'laps'                              : message_fields.get('num_laps'),
            'avg_hr'                            : message_fields.get('avg_heart_rate'),
            'max_hr'                            : message_fields.get('max_heart_rate'),
            'avg_rr'                            : message_fields.get('avg_respiration_rate'),
            'max_rr'                            : message_fields.get('max_respiration_rate'),
            'calories'                          : message_fields.get('total_calories'),
            'avg_cadence'                       : message_fields.get('avg_cadence'),
            'max_cadence'                       : message_fields.get('max_cadence'),
            'avg_speed'                         : message_fields.get('avg_speed'),
            'max_speed'                         : message_fields.get('max_speed'),
            'ascent'                            : message_fields.get('total_ascent'),
            'descent'                           : message_fields.get('total_descent'),
            'max_temperature'                   : message_fields.get('max_temperature'),
            'avg_temperature'                   : message_fields.get('avg_temperature'),
            'training_effect'                   : message_fields.get('total_training_effect'),
            'anaerobic_training_effect'         : message_fields.get('total_anaerobic_training_effect')
        }
        activity.update(self._plugin_dispatch('write_session_entry', self.garmin_act_db_session, fit_file, activity_id, message_fields))
        # json metadata gives better values for sport and subsport, so use existing value if set
        current = Activities.s_get(self.garmin_act_db_session, activity_id)
        if current:
            activity.update(self.__choose_sport(current.sport, current.sub_sport, sport, sub_sport))
            root_logger.debug("Updating with %r", activity)
            current.update_from_dict(activity, ignore_none=True, ignore_zero=True)
        else:
            activity.update({'sport': sport.name, 'sub_sport': sub_sport.name})
            root_logger.debug("Adding %r", activity)
            self.garmin_act_db_session.add(Activities(**activity))
        if sport is not None:
            function_name = '_write_' + sport.name + '_entry'
            try:
                function = getattr(self, function_name, None)
                if function is not None:
                    function(fit_file, activity_id, sub_sport, message_fields)
                else:
                    root_logger.warning("No sport handler for type %s from %s: %s", sport, fit_file.filename, message_fields)
            except Exception as e:
                root_logger.error("Exception in %s from %s: %s", function_name, fit_file.filename, e)

    def _write_hr_zones_timer_entry(self, fit_file, message_fields):
        """Write hz zones message to the database."""
        root_logger.info("writing hr zone data for %s", fit_file.filename)
        hr_zones_type = message_fields.get('hr_zones_timer_type')
        if hr_zones_type is fitfile.field_enums.HeartRateZonesTimerType.lap:
            self._write_hr_zones_timer_lap_entry(fit_file, message_fields)
        elif hr_zones_type is fitfile.field_enums.HeartRateZonesTimerType.session:
            self._write_hr_zones_timer_session_entry(fit_file, message_fields)

    def __hr_zone_data(self, message_fields):
        hr_zones_time = message_fields.get('hr_zones_time')
        zone_data = {
            'hr_zones_method'   : message_fields.get('hr_zones_method'),
            'hrz_1_time'        : hr_zones_time[1],
            'hrz_2_time'        : hr_zones_time[2],
            'hrz_3_time'        : hr_zones_time[3],
            'hrz_4_time'        : hr_zones_time[4],
            'hrz_5_time'        : hr_zones_time[5]
        }
        hr_zones = message_fields.get('hr_zones')
        if hr_zones:
            zone_data.update({
                'hrz_1_hr'      : hr_zones[0],
                'hrz_2_hr'      : hr_zones[1],
                'hrz_3_hr'      : hr_zones[2],
                'hrz_4_hr'      : hr_zones[3],
                'hrz_5_hr'      : hr_zones[4],
            })
        return zone_data

    def _write_hr_zones_timer_lap_entry(self, fit_file, message_fields):
        """Write lap hz zones message to the database."""
        root_logger.info("writing lap hr zone data %r for %s", message_fields, fit_file.filename)
        activity_id = File.id_from_path(fit_file.filename)
        lap = {
            'activity_id'   : activity_id,
            'lap'           : message_fields.get('record_num'),
        }
        lap.update(self.__hr_zone_data(message_fields))
        root_logger.info("writing lap hr zone data %r for %s", lap, fit_file.filename)
        ActivityLaps.s_insert_or_update(self.garmin_act_db_session, lap, ignore_none=True, ignore_zero=True)

    def _write_hr_zones_timer_session_entry(self, fit_file, message_fields):
        """Write session hz zones message to the database."""
        root_logger.info("writing session hr zone data %r for %s", message_fields, fit_file.filename)
        activity_id = File.id_from_path(fit_file.filename)
        session = {
            'activity_id'   : activity_id,
        }
        session.update(self.__hr_zone_data(message_fields))
        root_logger.debug("writing session hr zone data %r for %s", session, fit_file.filename)
        Activities.s_insert_or_update(self.garmin_act_db_session, session, ignore_none=True, ignore_zero=True)



"""Objects for importing Garmin activity data from Tcx files."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import sys
import logging
from tqdm import tqdm
import traceback

from idbutils import FileProcessor
from .tcx import Tcx

from .garmindb import GarminDb, Device, File, ActivitiesDb, Activities, ActivityRecords, ActivityLaps


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class GarminTcxData():
    """Class for importing Garmin activity data from TCX files."""

    def __init__(self, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminTcxData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        logger.info("Processing activities tcx data")
        self.measurement_system = measurement_system
        self.debug = debug
        if input_dir:
            self.file_names = FileProcessor.dir_to_files(input_dir, Tcx.filename_regex, latest)

    def file_count(self):
        """Return the number of files that will be propcessed."""
        return len(self.file_names)

    def __process_record(self, tcx, activity_id, record_number, point):
        root_logger.debug("Processing record: %r (%d)", point, record_number)
        if not ActivityRecords.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'record' : record_number}):
            record = {
                'activity_id'                       : activity_id,
                'record'                            : record_number,
                'timestamp'                         : tcx.get_point_time(point),
                'hr'                                : tcx.get_point_hr(point),
                'altitude'                          : tcx.get_point_altitude(point).meters_or_feet(measurement_system=self.measurement_system),
                'speed'                             : tcx.get_point_speed(point).kph_or_mph(measurement_system=self.measurement_system)
            }
            loc = tcx.get_point_loc(point)
            if loc is not None:
                record.update({'position_lat': loc.lat_deg, 'position_long': loc.long_deg})
            self.garmin_act_db_session.add(ActivityRecords(**record))

    def __process_lap(self, tcx, activity_id, lap_number, lap):
        root_logger.info("Processing lap: %d", lap_number)
        for record_number, point in enumerate(tcx.get_lap_points(lap)):
            self.__process_record(tcx, activity_id, record_number, point)
        if not ActivityLaps.s_exists(self.garmin_act_db_session, {'activity_id' : activity_id, 'lap' : lap_number}):
            lap_data = {
                'activity_id'                       : activity_id,
                'lap'                               : lap_number,
                'start_time'                        : tcx.get_lap_start(lap),
                'stop_time'                         : tcx.get_lap_end(lap),
                'elapsed_time'                      : tcx.get_lap_duration(lap),
                'distance'                          : tcx.get_lap_distance(lap).meters_or_feet(measurement_system=self.measurement_system),
                'calories'                          : tcx.get_lap_calories(lap)
            }
            start_loc = tcx.get_lap_start_loc(lap)
            if start_loc is not None:
                lap_data.update({'start_lat': start_loc.lat_deg, 'start_long': start_loc.long_deg})
            end_loc = tcx.get_lap_end_loc(lap)
            if end_loc is not None:
                lap_data.update({'stop_lat': end_loc.lat_deg, 'stop_long': end_loc.long_deg})
            root_logger.info("Inserting lap: %r (%d): %r", lap, lap_number, lap_data)
            self.garmin_act_db_session.add(ActivityLaps(**lap_data))

    def __process_file(self, file_name):
        tcx = Tcx()
        tcx.read(file_name)
        start_time = tcx.start_time
        (manufacturer, product) = tcx.get_manufacturer_and_product()
        serial_number = tcx.serial_number
        device = {
            'serial_number'     : serial_number,
            'timestamp'         : start_time,
            'manufacturer'      : manufacturer,
            'product'           : product,
            'hardware_version'  : None,
        }
        Device.s_insert_or_update(self.garmin_db_session, device, ignore_none=True)
        root_logger.info("Processing file: %s for manufacturer %s product %s device %s", file_name, manufacturer, product, serial_number)
        (file_id, file_name) = File.name_and_id_from_path(file_name)
        file = {
            'id'            : file_id,
            'name'          : file_name,
            'type'          : File.FileType.tcx,
            'serial_number' : serial_number,
        }
        File.s_insert_or_update(self.garmin_db_session, file)
        activity = {
            'activity_id'               : file_id,
            'name'                      : file_id,
            'start_time'                : start_time,
            'stop_time'                 : tcx.end_time,
            'laps'                      : tcx.lap_count,
            'sport'                     : tcx.sport,
            'calories'                  : tcx.calories,
            'distance'                  : tcx.distance.kms_or_miles(self.measurement_system),
            'avg_hr'                    : tcx.hr_avg,
            'max_hr'                    : tcx.hr_max,
            'max_cadence'               : tcx.cadence_max,
            'avg_cadence'               : tcx.cadence_avg,
            'ascent'                    : tcx.ascent.meters_or_feet(self.measurement_system),
            'descent'                   : tcx.descent.meters_or_feet(self.measurement_system)
        }
        start_loc = tcx.start_loc
        if start_loc is not None:
            activity.update({'start_lat': start_loc.lat_deg, 'start_long': start_loc.long_deg})
        end_loc = tcx.end_loc
        if end_loc is not None:
            activity.update({'stop_lat': end_loc.lat_deg, 'stop_long': end_loc.long_deg})
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True, ignore_zero=True)
        for lap_number, lap in enumerate(tcx.laps):
            self.__process_lap(tcx, file_id, lap_number, lap)

    def process_files(self, db_params):
        """Import data from TCX files into the database."""
        garmin_db = GarminDb(db_params, self.debug - 1)
        garmin_act_db = ActivitiesDb(db_params, self.debug - 1)
        with garmin_db.managed_session() as self.garmin_db_session, garmin_act_db.managed_session() as self.garmin_act_db_session:
            for file_name in tqdm(self.file_names, unit='files'):
                try:
                    self.__process_file(file_name)
                except Exception as e:
                    logger.error('Failed to processes TCX file %s: %s', file_name, e)
                    root_logger.error('Failed to processes TCX file %s: %s', file_name, traceback.format_exc())



"""Class for importing Garmin activity from FIT files."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import fitfile

from .fit_data import FitData


class GarminActivitiesFitData(FitData):
    """Class for importing Garmin activity data from FIT files."""

    def __init__(self, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminActivitiesFitData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        super().__init__(input_dir, debug, latest, False, [fitfile.FileType.activity], measurement_system)



"""Class that takes a parsed monitoring FIT file object and imports it into a database."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging
import sys

import fitfile

from .garmindb import SleepEvents
from .fit_file_processor import FitFileProcessor


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class SleepFitFileProcessor(FitFileProcessor):
    """Class that takes a parsed sleep FIT file object and imports it into a database."""

    def write_file(self, fit_file):
        """Given a Fit File object, write all of its messages to the DB."""
        self.last_sleep_event = None
        self.last_sleep_level = None
        with self.garmin_db.managed_session() as self.garmin_db_session:
            self._write_message_types(fit_file, fit_file.message_types)

    def _write_sleep_level_entry(self, fit_file, message_fields):
        logger.debug("sleep level message: %r", message_fields)
        timestamp = fit_file.utc_datetime_to_local(message_fields.timestamp)
        sleep_level = message_fields.get('sleep_level')
        if sleep_level.value > fitfile.field_enums.SleepActivityLevel.unknown.value and self.last_sleep_event is not None and \
           (sleep_level is not fitfile.field_enums.SleepActivityLevel.awake or self.last_sleep_level is not fitfile.field_enums.SleepActivityLevel.awake):
            sleep_event = {
                'timestamp' : fit_file.utc_datetime_to_local(self.last_sleep_event),
                'event'     : sleep_level.name,
                'duration'  : fitfile.conversions.timedelta_to_time(timestamp - self.last_sleep_event)
            }
            logger.debug("sleep level event: %r", sleep_event)
            SleepEvents.s_insert_or_update(self.garmin_db_session, sleep_event)
        self.last_sleep_event = timestamp
        self.last_sleep_level = sleep_level



"""Base class for GarminDb plugins."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging

from .plugin_base import PluginBase


logger = logging.getLogger(__file__)


class ActivityFitPluginBase(PluginBase):
    """Base class for GarminDb activity FIT file plugins that handle data based on ids of applications or developer fields, sport or sub-sport ids, etc."""

    _type = 'ActivityFit'

    @classmethod
    def matches_activity_file(cls, fit_file):
        """Return if the file matches this plugin."""
        if hasattr(cls, '_application_id'):
            return cls._application_id in fit_file.dev_application_ids
        if hasattr(cls, '_sport') and (fit_file.sport_type is None or fit_file.sport_type.value is not cls._sport):
            return False
        if hasattr(cls, '_sub_sport') and (fit_file.sub_sport_type is None or fit_file.sub_sport_type.value is not cls._sub_sport):
            return False
        if hasattr(cls, '_dev_fields'):
            for dev_field in cls._dev_fields:
                if dev_field not in fit_file.dev_fields:
                    return False
        return True

    @classmethod
    def init_activity(cls, act_db_class, activities_table):
        """Initialize an instance of the plugin as an activity FIT file plugin."""
        logger.info("Initializing tables for activity plugin %s with activities table %s", cls.__name__, activities_table)
        if hasattr(cls, '_records_tablename') and 'record' not in cls._tables:
            cls._tables['record'] = activities_table.create(cls._records_tablename, act_db_class, cls._records_version, cls._records_pk, cls._records_cols)
        if hasattr(cls, '_laps_tablename') and 'lap' not in cls._tables:
            cls._tables['lap'] = activities_table.create(cls._laps_tablename, act_db_class, cls._laps_version, cls._laps_pk, cls._laps_cols)
        if hasattr(cls, '_sessions_tablename') and 'session' not in cls._tables:
            cls._tables['session'] = activities_table.create(cls._sessions_tablename, act_db_class, cls._sessions_version, cols=cls._sessions_cols,
                                                             create_view=cls._views['activity_view'], vars={'activities_table': activities_table})



"""Base class for GarminDb plugins."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import logging


logger = logging.getLogger(__file__)


class PluginBase():
    """Base class for GarminDb file plugins."""

    @classmethod
    def _get_field(cls, message_fields, field_name_list):
        for field_name in field_name_list:
            if field_name in message_fields:
                return message_fields[field_name]

    @classmethod
    def filter_data(cls, indict):
        """Filter None and 0 values from a dict."""
        return {key: value for key, value in indict.items() if value}

    def __str__(self):
        """Return a string representation of the class instance."""
        return f'{self.__class__.__name__}(tables {repr(self._tables.keys())})'



"""Test config handling."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os
import unittest
import logging
import sys

from garmindb import GarminConnectConfigManager


root_logger = logging.getLogger()
handler = logging.FileHandler('copy.log', 'w')
root_logger.addHandler(handler)
root_logger.setLevel(logging.INFO)

logger = logging.getLogger(__name__)


class TestConfig(unittest.TestCase):
    """Class for testing config handling."""

    @classmethod
    def setUpClass(cls):
        cls.gc_config = GarminConnectConfigManager()
        cls.homedir = os.path.expanduser('~')

    def test_directories(self):
        # config_dir
        expected_config_dir = self.homedir + os.sep + '.GarminDb'
        config_dir = self.gc_config.config_dir
        self.assertEqual(config_dir, expected_config_dir, f'actual {config_dir} expected {expected_config_dir}')
        # base_dir
        expected_base_dir = self.homedir + os.sep + 'HealthData'
        base_dir = self.gc_config.get_base_dir()
        self.assertEqual(base_dir, expected_base_dir, f'actual {base_dir} expected {expected_base_dir}')
        # monitoring_dir
        year = 2023
        expected_monitoring_dir = expected_base_dir + os.sep + "FitFiles" + os.sep + 'Monitoring' + os.sep + str(year)
        monitoring_dir = self.gc_config.get_monitoring_dir(year)
        self.assertEqual(monitoring_dir, expected_monitoring_dir, f'actual {monitoring_dir} expected {expected_monitoring_dir}')

    def test_db(self):
        db_params = self.gc_config.get_db_params()
        expect_db_type = 'sqlite'
        self.assertEqual(db_params.db_type, expect_db_type, f"expected {expect_db_type} actual {db_params.db_type}")
        expected_db_path = self.homedir + os.sep + 'HealthData' + os.sep + 'DBs'
        self.assertEqual(db_params.db_path, expected_db_path, f"expected {expected_db_path} actual {db_params.db_path}")

if __name__ == '__main__':
    unittest.main(verbosity=2)


"""Objects for analyzing health data from Garmin devices and genrating summary tables."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import logging
import datetime
import calendar
from tqdm import tqdm

import fitfile

from garmindb import summarydb
from .garmindb import GarminDb, Attributes, Weight, Stress, RestingHeartRate, IntensityHR, Sleep
from .garmindb import MonitoringDb, Monitoring, MonitoringHeartRate, MonitoringIntensity, MonitoringClimb
from .garmindb import ActivitiesDb, Activities, StepsActivities
from .garmindb import GarminSummaryDb, DaysSummary, DailySummary, WeeksSummary, MonthsSummary, YearsSummary


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))


class Analyze():
    """Object for analyzing health data from Garmin devices."""

    def __init__(self, gc_config, debug):
        """Return an instance of the Analyze class."""
        self.gc_config = gc_config
        self.garmin_db = GarminDb(self.gc_config.get_db_params(), debug)
        self.garmin_mon_db = MonitoringDb(self.gc_config.get_db_params(), debug)
        self.garmin_sum_db = GarminSummaryDb(self.gc_config.get_db_params(), debug)
        self.sum_db = summarydb.SummaryDb(self.gc_config.get_db_params(), debug)
        self.garmin_act_db = ActivitiesDb(self.gc_config.get_db_params(), debug)
        self.measurement_system = Attributes.measurements_type(self.garmin_db)
        self.unit_strings = fitfile.units.unit_strings[self.measurement_system]

    def __populate_hr_intensity(self, day_date, garmin_mon_session, garmin_sum_session, overwrite=False):
        if IntensityHR.s_row_count_for_day(garmin_sum_session, day_date) == 0 or overwrite:
            monitoring_rows = Monitoring._get_for_day(garmin_mon_session, day_date, not_none_col=Monitoring.intensity)
            previous_ts = None
            for monitoring in monitoring_rows:
                # Heart rate value is for one minute, reported at the end of the minute. Only take HR values where the
                # measurement period falls within the activity period.
                if previous_ts is not None and (monitoring.timestamp - previous_ts).total_seconds() > 60:
                    hr_rows = MonitoringHeartRate.s_get_for_period(garmin_mon_session, previous_ts, previous_ts + datetime.timedelta(seconds=60), not_none_col=monitoring.timestamp)
                    for hr in hr_rows:
                        entry = {
                            'timestamp'     : hr.timestamp,
                            'intensity'     : monitoring.intensity,
                            'heart_rate'    : hr.heart_rate
                        }
                        IntensityHR.s_insert_or_update(garmin_sum_session, entry, ignore_none=True)
                previous_ts = monitoring.timestamp

    def __calculate_day_stats(self, day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        stats = DailySummary.get_daily_stats(garmin_session, day_date)
        # prefer getting stats from the daily summary.
        if stats.get('rhr_avg') is None:
            stats.update(RestingHeartRate.get_daily_stats(garmin_session, day_date))
        if stats.get('stress_avg') is None:
            stats.update(Stress.get_daily_stats(garmin_session, day_date))
        if stats.get('intensity_time') is None:
            stats.update(MonitoringIntensity.get_daily_stats(garmin_mon_session, day_date))
        if stats.get('floors') is None:
            stats.update(MonitoringClimb.get_daily_stats(garmin_mon_session, day_date, self.measurement_system))
        if stats.get('steps') is None:
            stats.update(Monitoring.get_daily_stats(garmin_mon_session, day_date))
        stats.update(MonitoringHeartRate.get_daily_stats(garmin_mon_session, day_date))
        stats.update(IntensityHR.get_daily_stats(garmin_sum_session, day_date))
        stats.update(Weight.get_daily_stats(garmin_session, day_date))
        stats.update(Sleep.get_daily_stats(garmin_session, day_date))
        # save it to the db
        DaysSummary.s_insert_or_update(garmin_sum_session, stats)
        summarydb.DaysSummary.s_insert_or_update(sum_session, stats)

    def __calculate_days(self, year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        days = Monitoring.s_get_days(garmin_mon_session, year)
        if days:
            for day in tqdm(days, unit='days'):
                day_date = datetime.date(year, 1, 1) + datetime.timedelta(day - 1)
                self.__populate_hr_intensity(day_date, garmin_mon_session, garmin_sum_session)
                self.__calculate_day_stats(day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)
        days = Activities.s_get_days(garmin_act_session, year)
        if len(days):
            for day in tqdm(days, unit='days'):
                stats = Activities.get_daily_stats(garmin_act_session, datetime.date(year, 1, 1) + datetime.timedelta(day - 1))
                DaysSummary.s_insert_or_update(garmin_sum_session, stats)
                summarydb.DaysSummary.s_insert_or_update(sum_session, stats)

    def __calculate_week_stats(self, day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        stats = DailySummary.get_weekly_stats(garmin_session, day_date)
        # prefer getting stats from the daily summary.
        if stats.get('rhr_avg') is None:
            stats.update(RestingHeartRate.get_weekly_stats(garmin_session, day_date))
        if stats.get('stress_avg') is None:
            stats.update(Stress.get_weekly_stats(garmin_session, day_date))
        if stats.get('intensity_time') is None:
            stats.update(MonitoringIntensity.get_weekly_stats(garmin_mon_session, day_date))
        if stats.get('floors') is None:
            stats.update(MonitoringClimb.get_weekly_stats(garmin_mon_session, day_date, self.measurement_system))
        if stats.get('steps') is None:
            stats.update(Monitoring.get_weekly_stats(garmin_mon_session, day_date))
        stats.update(MonitoringHeartRate.get_weekly_stats(garmin_mon_session, day_date))
        stats.update(IntensityHR.get_weekly_stats(garmin_sum_session, day_date))
        stats.update(Weight.get_weekly_stats(garmin_session, day_date))
        stats.update(Sleep.get_weekly_stats(garmin_session, day_date))
        stats.update(Activities.get_weekly_stats(garmin_act_session, day_date))
        # save it to the db
        WeeksSummary.s_insert_or_update(garmin_sum_session, stats)
        summarydb.WeeksSummary.s_insert_or_update(sum_session, stats)

    def __calculate_weeks(self, year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        for week_starting_day in tqdm(range(1, 365, 7), unit='weeks'):
            day_date = datetime.date(year, 1, 1) + datetime.timedelta(week_starting_day - 1)
            if day_date < datetime.datetime.now().date():
                self.__calculate_week_stats(day_date, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)

    def __calculate_monitoring_month_stats(self, start_day_date, end_day_date, garmin_session, garmin_mon_session, garmin_sum_session, sum_session):
        stats = DailySummary.get_monthly_stats(garmin_session, start_day_date, end_day_date)
        # prefer getting stats from the daily summary.
        if 'rhr_avg' in stats:
            stats.update(RestingHeartRate.get_monthly_stats(garmin_session, start_day_date, end_day_date))
        if 'stress_avg' in stats:
            stats.update(Stress.get_monthly_stats(garmin_session, start_day_date, end_day_date))
        if 'intensity_time' in stats:
            stats.update(MonitoringIntensity.get_monthly_stats(garmin_mon_session, start_day_date, end_day_date))
        if 'floors' in stats:
            stats.update(MonitoringClimb.get_monthly_stats(garmin_mon_session, start_day_date, end_day_date, self.measurement_system))
        if 'steps' in stats:
            stats.update(Monitoring.get_monthly_stats(garmin_mon_session, start_day_date, end_day_date))
        stats.update(MonitoringHeartRate.get_monthly_stats(garmin_mon_session, start_day_date, end_day_date))
        stats.update(IntensityHR.get_monthly_stats(garmin_sum_session, start_day_date, end_day_date))
        stats.update(Weight.get_monthly_stats(garmin_session, start_day_date, end_day_date))
        stats.update(Sleep.get_monthly_stats(garmin_session, start_day_date, end_day_date))
        # save it to the db
        MonthsSummary.s_insert_or_update(garmin_sum_session, stats)
        summarydb.MonthsSummary.s_insert_or_update(sum_session, stats)

    def __calculate_months(self, year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        months = Monitoring.s_get_months(garmin_mon_session, year)
        if len(months):
            for month in tqdm(months, unit='months'):
                start_day_date = datetime.date(year, month, 1)
                end_day_date = datetime.date(year, month, calendar.monthrange(year, month)[1])
                self.__calculate_monitoring_month_stats(start_day_date, end_day_date, garmin_session, garmin_mon_session, garmin_sum_session, sum_session)
        months = Activities.s_get_months(garmin_act_session, year)
        if len(months):
            for month in tqdm(months, unit='months'):
                stats = Activities.get_monthly_stats(garmin_act_session, datetime.date(year, month, 1), datetime.date(year, month, calendar.monthrange(year, month)[1]))
                MonthsSummary.s_insert_or_update(garmin_sum_session, stats)
                summarydb.MonthsSummary.s_insert_or_update(sum_session, stats)

    def __calculate_year_stats(self, year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session):
        stats = DailySummary.get_yearly_stats(garmin_session, year)
        # prefer getting stats from the daily summary.
        if 'rhr_avg' in stats:
            stats.update(RestingHeartRate.get_yearly_stats(garmin_session, year))
        if 'stress_avg' in stats:
            stats.update(Stress.get_yearly_stats(garmin_session, year))
        if 'intensity_time' in stats:
            stats.update(MonitoringIntensity.get_yearly_stats(garmin_mon_session, year))
        if 'floors' in stats:
            stats.update(MonitoringClimb.get_yearly_stats(garmin_mon_session, year, self.measurement_system))
        if 'steps' in stats:
            stats.update(Monitoring.get_yearly_stats(garmin_mon_session, year))
        stats.update(MonitoringHeartRate.get_yearly_stats(garmin_mon_session, year))
        stats.update(IntensityHR.get_yearly_stats(garmin_sum_session, year))
        stats.update(Weight.get_yearly_stats(garmin_session, year))
        stats.update(Sleep.get_yearly_stats(garmin_session, year))
        stats.update(Activities.get_yearly_stats(garmin_act_session, year))
        # save it to the db
        YearsSummary.s_insert_or_update(garmin_sum_session, stats)
        summarydb.YearsSummary.s_insert_or_update(sum_session, stats)

    def __calculate_year(self, year):
        with self.garmin_db.managed_session() as garmin_session, self.garmin_mon_db.managed_session() as garmin_mon_session, \
                self.garmin_act_db.managed_session() as garmin_act_session, self.garmin_sum_db.managed_session() as garmin_sum_session, \
                self.sum_db.managed_session() as sum_session:
            # calculate part of the years
            self.__calculate_days(year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)
            self.__calculate_weeks(year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)
            self.__calculate_months(year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)
            # now calculate the year itself
            self.__calculate_year_stats(year, garmin_session, garmin_mon_session, garmin_act_session, garmin_sum_session, sum_session)

    def summary(self):
        """Summarize Garmin health data. Daily, weekly, and monthly, tables will be generated."""
        logger.info("Summary Tables Generation:")
        for year in sorted(list(set(Monitoring.get_years(self.garmin_mon_db) + Activities.get_years(self.garmin_act_db)))):
            logger.info("Generating table entries for %s", year)
            self.__calculate_year(year)

    def create_dynamic_views(self):
        """Create database views specific to the data in this database."""
        course_ids = self.gc_config.course_views('steps')
        if course_ids:
            for course_id in course_ids:
                StepsActivities.create_course_view(self.garmin_act_db, course_id)



"""Class for downloading health data from Garmin Connect."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import os
import sys
import re
import logging
import datetime
import time
import tempfile
import zipfile
import json
from garth import Client as GarthClient
from garth.exc import GarthHTTPError, GarthException
from tqdm import tqdm

import fitfile.conversions as conversions


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class Download():
    """Class for downloading health data from Garmin Connect."""

    garmin_connect_user_profile_url = "/userprofile-service/userprofile"
    garmin_connect_wellness_url = "/wellness-service/wellness"
    garmin_connect_sleep_daily_url = garmin_connect_wellness_url + "/dailySleepData"
    garmin_connect_rhr = "/userstats-service/wellness/daily"
    garmin_connect_weight_url = "/weight-service/weight/dateRange"

    garmin_connect_activity_search_url = "/activitylist-service/activities/search/activities"
    garmin_connect_activity_service_url = "/activity-service/activity"

    garmin_connect_download_service_url = "/download-service/files"

    garmin_connect_usersummary_url = "/usersummary-service/usersummary"
    garmin_connect_daily_summary_url = garmin_connect_usersummary_url + "/daily"
    garmin_connect_daily_hydration_url = garmin_connect_usersummary_url + "/hydration/allData"

    # https://connect.garmin.com/modern/proxy/usersummary-service/usersummary/hydration/allData/2019-11-29

    download_days_overlap = 3  # Existing donloaded data will be redownloaded and overwritten if it is within this number of days of now.

    def __init__(self, gc_config):
        """Create a new Download class instance."""
        logger.debug("__init__")
        self.gc_config = gc_config
        self.garth_session_file = self.gc_config.get_session_file()
        self.garth = GarthClient()
        self.garth.configure(domain=self.gc_config.get_garmin_base_domain())

    def __resume_session(self):
        if os.path.isfile(self.garth_session_file):
            root_logger.info("load session from %s", self.garth_session_file)
            with open(self.garth_session_file, "r", encoding="utf-8") as file:
                self.garth.loads(file.read())
                return True
        else:
            root_logger.info("session file %s not found", self.garth_session_file)
        return False

    def __save_session(self):
        root_logger.info("save session to %s", self.garth_session_file)
        with open(self.garth_session_file, "w", encoding="utf-8") as file:
            file.write(self.garth.dumps())

    def __login(self):
        username = self.gc_config.get_user()
        password = self.gc_config.get_password()
        if not username or not password:
            print("Missing config: need username and password. Edit GarminConnectConfig.json.")
            return

        logger.debug("login: %s %s", username, password)
        self.garth.login(username, password)
        self.__save_session()

    def login(self):
        """Use garth to resume to Garmin Connect session if possible, otherwise login."""
        if not self.__resume_session():
            self.__login()

        try:
            self.garth.username
        except GarthException:
            self.__login()

        profile_dir = self.gc_config.get_fit_files_dir()
        self.save_json_to_file(f'{profile_dir}/social-profile', self.garth.profile)
        self.save_json_to_file(f'{profile_dir}/user-settings', self.garth.connectapi(f'{self.garmin_connect_user_profile_url}/user-settings'), True)
        self.save_json_to_file(f'{profile_dir}/personal-information', self.garth.connectapi(f'{self.garmin_connect_user_profile_url}/personal-information'), True)

        self.display_name = self.garth.profile['displayName']
        self.full_name = self.garth.profile['fullName']
        root_logger.info("login: %s (%s)", self.full_name, self.display_name)
        return True

    def __unzip_files(self, outdir):
        """Unzip and downloaded zipped files into the directory supplied."""
        root_logger.info("unzip_files: from %s to %s", self.temp_dir, outdir)
        for filename in os.listdir(self.temp_dir):
            match = re.search(r'.*\.zip', filename)
            if match:
                full_pathname = f'{self.temp_dir}/{filename}'
                with zipfile.ZipFile(full_pathname, 'r') as files_zip:
                    try:
                        files_zip.extractall(outdir)
                    except Exception as e:
                        logger.error('Failed to unzip %s to %s: %s', full_pathname, outdir, e)

    @classmethod
    def __convert_to_json(cls, object):
        return object.__str__()

    @classmethod
    def save_json_to_file(cls, filename, json_data, overwrite=False):
        """Save JSON formatted data to a file."""
        full_filename = f'{filename}.json'
        exists = os.path.isfile(full_filename)
        if not exists or overwrite:
            logger.debug("%s %s", 'Overwriting' if exists else 'Saving', full_filename)
            with open(full_filename, 'w') as file:
                file.write(json.dumps(json_data, default=cls.__convert_to_json))

    def save_binary_file(self, filename, url, overwrite=False):
        """Save binary data to a file."""
        exists = os.path.isfile(filename)
        if not exists or overwrite:
            logger.debug("%s %s", 'Overwriting' if exists else 'Saving', filename)
            response = self.garth.get("connectapi", url, api=True)
            with open(filename, 'wb') as file:
                for chunk in response:
                    file.write(chunk)

    def __get_stat(self, stat_function, directory, date, days, overwrite):
        for day in tqdm(range(0, days), unit='days'):
            download_date = date + datetime.timedelta(days=day)
            # always overwrite for yesterday and today since the last download may have been a partial result
            delta = datetime.datetime.now().date() - download_date
            stat_function(directory, download_date, overwrite or delta.days <= self.download_days_overlap)
            # pause for a second between every page access
            time.sleep(1)

    def __get_summary_day(self, directory_func, date, overwrite=False):
        root_logger.info("get_summary_day: %s", date)
        date_str = date.strftime('%Y-%m-%d')
        params = {
            'calendarDate': date_str,
            '_': str(conversions.dt_to_epoch_ms(conversions.date_to_dt(date)))
        }
        url = f'{self.garmin_connect_daily_summary_url}/{self.display_name}'
        json_filename = f'{directory_func(date.year)}/daily_summary_{date_str}'
        try:
            self.save_json_to_file(json_filename, self.garth.connectapi(url, params=params), overwrite)
        except GarthHTTPError as e:
            root_logger.error("Exception getting daily summary: %s", e)

    def get_daily_summaries(self, directory_func, date, days, overwrite):
        """Download the daily summary data from Garmin Connect and save to a JSON file."""
        root_logger.info("Getting daily summaries: %s (%d)", date, days)
        self.__get_stat(self.__get_summary_day, directory_func, date, days, overwrite)

    def __get_monitoring_day(self, date):
        root_logger.info("get_monitoring_day: %s to %s", date, self.temp_dir)
        zip_filename = f'{self.temp_dir}/{date}.zip'
        url = f'{self.garmin_connect_download_service_url}/wellness/{date.strftime("%Y-%m-%d")}'
        try:
            self.save_binary_file(zip_filename, url)
        except GarthHTTPError as e:
            root_logger.error("Exception getting daily summary: %s", e)

    def get_monitoring(self, directory_func, date, days):
        """Download the daily monitoring data from Garmin Connect, unzip and save the raw files."""
        root_logger.info("Getting monitoring: %s (%d)", date, days)
        for day in tqdm(range(0, days), unit='days'):
            day_date = date + datetime.timedelta(day)
            self.temp_dir = tempfile.mkdtemp()
            self.__get_monitoring_day(day_date)
            self.__unzip_files(directory_func(day_date.year))
            # pause for a second between every page access
            time.sleep(1)

    def __get_weight_day(self, directory, day, overwrite=False):
        root_logger.info("Checking weight: %s overwrite %r", day, overwrite)
        date_str = day.strftime('%Y-%m-%d')
        params = {
            'startDate' : date_str,
            'endDate'   : date_str,
            '_'         : str(conversions.dt_to_epoch_ms(conversions.date_to_dt(day)))
        }
        json_filename = f'{directory}/weight_{date_str}'
        try:
            self.save_json_to_file(json_filename, self.garth.connectapi(self.garmin_connect_weight_url, params=params), overwrite)
        except GarthHTTPError as e:
            root_logger.error("Exception getting daily summary: %s", e)

    def get_weight(self, directory, date, days, overwrite):
        """Download the sleep data from Garmin Connect and save to a JSON file."""
        root_logger.info("Getting weight: %s (%d)", date, days)
        self.__get_stat(self.__get_weight_day,
                        directory, date, days, overwrite)

    def __get_activity_summaries(self, start, count):
        root_logger.info("get_activity_summaries")
        params = {
            'start' : str(start),
            "limit" : str(count)
        }
        try:
            return self.garth.connectapi(self.garmin_connect_activity_search_url, params=params)
        except GarthHTTPError as e:
            root_logger.error("Exception getting activity summary: %s", e)

    def __save_activity_details(self, directory, activity_id_str, overwrite):
        root_logger.debug("save_activity_details")
        json_filename = f'{directory}/activity_details_{activity_id_str}'
        try:
            url = f'{self.garmin_connect_activity_service_url}/{activity_id_str}'
            self.save_json_to_file(
                json_filename, self.garth.connectapi(url), overwrite)
        except GarthHTTPError as e:
            root_logger.error("Exception getting daily summary %s", e)

    def __save_activity_file(self, activity_id_str):
        root_logger.debug("save_activity_file: %s", activity_id_str)
        zip_filename = f'{self.temp_dir}/activity_{activity_id_str}.zip'
        url = f'{self.garmin_connect_download_service_url}/activity/{activity_id_str}'
        try:
            self.save_binary_file(zip_filename, url)
        except GarthHTTPError as e:
            root_logger.error("Exception downloading activity file: %s", e)

    def get_activities(self, directory, count, overwrite=False):
        """Download activities files from Garmin Connect and save the raw files."""
        self.temp_dir = tempfile.mkdtemp()
        logger.info("Getting activities: '%s' (%d) temp %s", directory, count, self.temp_dir)
        activities = self.__get_activity_summaries(0, count)
        for activity in tqdm(activities or [], unit='activities'):
            activity_id_str = str(activity['activityId'])
            activity_name_str = conversions.printable(activity.get('activityName'))
            root_logger.info("get_activities: %s (%s)", activity_name_str, activity_id_str)
            json_filename = f'{directory}/activity_{activity_id_str}'
            if not os.path.isfile(json_filename + '.json') or overwrite:
                root_logger.info("get_activities: %s <- %r", json_filename, activity)
                self.__save_activity_details(directory, activity_id_str, overwrite)
                self.save_json_to_file(json_filename, activity)
                if not os.path.isfile(f'{directory}/{activity_id_str}.fit') or overwrite:
                    self.__save_activity_file(activity_id_str)
                # pause for a second between every page access
                time.sleep(1)
            else:
                root_logger.info("get_activities: skipping download of %s, already present", activity_id_str)
        self.__unzip_files(directory)

    def get_activity_types(self, directory, overwrite):
        """Download the activity types from Garmin Connect and save to a JSON file."""
        root_logger.info("get_activity_types: '%s'", directory)
        json_filename = f'{directory}/activity_types'
        try:
            url = f'{self.garmin_connect_activity_service_url}/activityTypes'
            self.save_json_to_file(
                json_filename, self.garth.connectapi(url), overwrite)
        except GarthHTTPError as e:
            root_logger.error("Exception getting activity types: %s", e)

    def __get_sleep_day(self, directory, date, overwrite=False):
        json_filename = f'{directory}/sleep_{date}'
        params = {
            'date'                  : date.strftime("%Y-%m-%d"),
            'nonSleepBufferMinutes' : 60
        }
        url = f'{self.garmin_connect_sleep_daily_url}/{self.display_name}'
        try:
            self.save_json_to_file(json_filename, self.garth.connectapi(url, params=params), overwrite)
        except GarthHTTPError as e:
            root_logger.error("Exception getting daily summary: %s", e)

    def get_sleep(self, directory, date, days, overwrite):
        """Download the sleep data from Garmin Connect and save to a JSON file."""
        root_logger.info("Getting sleep: %s (%d)", date, days)
        self.__get_stat(self.__get_sleep_day, directory, date, days, overwrite)

    def __get_rhr_day(self, directory, day, overwrite=False):
        date_str = day.strftime('%Y-%m-%d')
        json_filename = f'{directory}/rhr_{date_str}'
        params = {
            'fromDate'  : date_str,
            'untilDate' : date_str,
            'metricId'  : 60
        }
        url = f'{self.garmin_connect_rhr}/{self.display_name}'
        try:
            self.save_json_to_file(json_filename, self.garth.connectapi(
                url, params=params), overwrite)
        except GarthHTTPError as e:
            root_logger.error("Exception getting daily summary %s", e)

    def get_rhr(self, directory, date, days, overwrite):
        """Download the resting heart rate data from Garmin Connect and save to a JSON file."""
        root_logger.info("Getting rhr: %s (%d)", date, days)
        self.__get_stat(self.__get_rhr_day, directory, date, days, overwrite)

    def __get_hydration_day(self, directory_func, day, overwrite=False):
        date_str = day.strftime('%Y-%m-%d')
        json_filename = f'{directory_func(day.year)}/hydration_{date_str}'
        url = f'{self.garmin_connect_daily_hydration_url}/{date_str}'
        try:
            self.save_json_to_file(json_filename, self.garth.connectapi(url), overwrite)
        except GarthHTTPError as e:
            root_logger.error("Exception getting hydration: %s", e)

    def get_hydration(self, directory_func, date, days, overwrite):
        """Download the hydration data from Garmin Connect and save to a JSON file."""
        root_logger.info("Getting hydration: %s (%d)", date, days)
        self.__get_stat(self.__get_hydration_day, directory_func, date, days, overwrite)



"""Objects for importing Garmin activity data from Garmin Connect downloads and FIT files."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"


import sys
import logging
import dateutil.parser

import fitfile
from idbutils import JsonFileProcessor

from .garmin_connect_enums import Event, get_summary_sport, get_details_sport
from .garmindb import ActivitiesDb, Activities, StepsActivities, PaddleActivities, CycleActivities


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class GarminJsonActivityData(JsonFileProcessor):
    """Base class for importing Garmin activity data from JSON formatted Garmin Connect details downloads."""

    def __init__(self, db_params, file_regex, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminJsonDetailsData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        super().__init__(file_regex, input_dir=input_dir, latest=latest, debug=debug)
        self.measurement_system = measurement_system
        self.garmin_act_db = ActivitiesDb(db_params, self.debug - 1)
        self.conversions = {}

    def _process_common(self, json_data):
        distance = self._get_field_obj(json_data, 'distance', fitfile.Distance.from_meters)
        ascent = self._get_field_obj(json_data, 'elevationGain', fitfile.Distance.from_meters)
        descent = self._get_field_obj(json_data, 'elevationLoss', fitfile.Distance.from_meters)
        avg_speed = self._get_field_obj(json_data, 'averageSpeed', fitfile.Speed.from_mps)
        max_speed = self._get_field_obj(json_data, 'maxSpeed', fitfile.Speed.from_mps)
        max_temperature = self._get_field_obj(json_data, 'maxTemperature', fitfile.Temperature.from_celsius)
        min_temperature = self._get_field_obj(json_data, 'minTemperature', fitfile.Temperature.from_celsius)
        avg_temperature = self._get_field_obj(json_data, 'averageTemperature', fitfile.Temperature.from_celsius)
        start_time = dateutil.parser.parse(self._get_field(json_data, 'startTimeLocal'), ignoretz=True)
        elapsed_time = fitfile.conversions.secs_to_dt_time(self._get_field(json_data, 'elapsedDuration', int))
        return {
            'start_time'                : start_time,
            'stop_time'                 : start_time + fitfile.conversions.time_to_timedelta(elapsed_time) if elapsed_time is not None else None,
            'elapsed_time'              : elapsed_time,
            'moving_time'               : fitfile.conversions.secs_to_dt_time(self._get_field(json_data, 'movingDuration', int)),
            'start_lat'                 : self._get_field(json_data, 'startLatitude', float),
            'start_long'                : self._get_field(json_data, 'startLongitude', float),
            'stop_lat'                  : self._get_field(json_data, 'endLatitude', float),
            'stop_long'                 : self._get_field(json_data, 'endLongitude', float),
            'distance'                  : distance.kms_or_miles(self.measurement_system),
            'laps'                      : self._get_field(json_data, 'lapCount'),
            'avg_hr'                    : self._get_field(json_data, 'averageHR', float),
            'max_hr'                    : self._get_field(json_data, 'maxHR', float),
            'calories'                  : self._get_field(json_data, 'calories', float),
            'avg_speed'                 : avg_speed.kph_or_mph(self.measurement_system) if avg_speed is not None else None,
            'max_speed'                 : max_speed.kph_or_mph(self.measurement_system) if max_speed is not None else None,
            'ascent'                    : ascent.meters_or_feet(self.measurement_system) if ascent is not None else None,
            'descent'                   : descent.meters_or_feet(self.measurement_system) if descent is not None else None,
            'max_temperature'           : max_temperature.c_or_f(self.measurement_system) if max_temperature is not None else None,
            'min_temperature'           : min_temperature.c_or_f(self.measurement_system) if min_temperature is not None else None,
            'avg_temperature'           : avg_temperature.c_or_f(self.measurement_system) if avg_temperature is not None else None,
            'training_effect'           : self._get_field(json_data, 'aerobicTrainingEffect', float),
            'anaerobic_training_effect' : self._get_field(json_data, 'anaerobicTrainingEffect', float),
            'max_rr'                    : self._get_field(json_data, 'maxRespirationRate', float),
            'avg_rr'                    : self._get_field(json_data, 'avgRespirationRate', float),
        }

    def _process_json(self, json_data):
        """Import data from files into the database."""
        with self.garmin_act_db.managed_session() as self.garmin_act_db_session:
            return self._activities_process_json(json_data)


class GarminJsonSummaryData(GarminJsonActivityData):
    """Class for importing Garmin activity data from JSON formatted Garmin Connect summary downloads."""

    def __init__(self, db_params, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminTcxData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        logger.info("Processing %s activities summary data from %s", 'latest' if latest else 'all', input_dir)
        super().__init__(db_params, r'activity_\d*\.json', input_dir, latest, measurement_system, debug)

    def _process_steps_activity(self, activity_id, activity_summary):
        root_logger.debug("process_steps_activity for %s", activity_id)
        avg_vertical_oscillation = self._get_field_obj(activity_summary, 'avgVerticalOscillation', fitfile.Distance.from_meters)
        avg_step_length = self._get_field_obj(activity_summary, 'avgStrideLength', fitfile.Distance.from_meters)
        run = {
            'activity_id'               : activity_id,
            'steps'                     : self._get_field(activity_summary, 'steps', float),
            'avg_steps_per_min'         : self._get_field(activity_summary, 'averageRunningCadenceInStepsPerMinute', float),
            'max_steps_per_min'         : self._get_field(activity_summary, 'maxRunningCadenceInStepsPerMinute', float),
            'avg_step_length'           : avg_step_length.meters_or_feet(self.measurement_system),
            'avg_gct_balance'           : self._get_field(activity_summary, 'avgGroundContactBalance', float),
            'avg_vertical_oscillation'  : avg_vertical_oscillation.meters_or_feet(self.measurement_system),
            'avg_ground_contact_time'   : fitfile.conversions.ms_to_dt_time(self._get_field(activity_summary, 'avgGroundContactTime', float)),
            'vo2_max'                   : self._get_field(activity_summary, 'vO2MaxValue', float),
        }
        StepsActivities.s_insert_or_update(self.garmin_act_db_session, run, ignore_none=True)

    def _process_inline_skating(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("inline_skating for %s: %r", activity_id, activity_summary)

    def _process_snowshoeing(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("snow_shoe for %s: %r", activity_id, activity_summary)

    def _process_strength_training(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("strength_training for %s: %r", activity_id, activity_summary)

    def _process_stand_up_paddleboarding(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("stand_up_paddleboarding for %s: %r", activity_id, activity_summary)

    def _process_resort_skiing_snowboarding(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("resort_skiing_snowboarding for %s: %r", activity_id, activity_summary)

    def _process_running(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("process_running for %s", activity_id)
        self._process_steps_activity(activity_id, activity_summary)
    #
    # def _process_treadmill_running(self, sub_sport, activity_id, activity_summary):
    #     root_logger.debug("process_treadmill_running for %s", activity_id)
    #     self._process_steps_activity(activity_id, activity_summary)

    def _process_walking(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("process_walking for %s", activity_id)
        self._process_steps_activity(activity_id, activity_summary)

    def _process_hiking(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("process_hiking for %s", activity_id)
        self._process_steps_activity(activity_id, activity_summary)

    def _process_paddling(self, sub_sport, activity_id, activity_summary):
        activity = {
            'activity_id'               : activity_id,
            'avg_cadence'               : self._get_field(activity_summary, 'avgStrokeCadence', float),
            'max_cadence'               : self._get_field(activity_summary, 'maxStrokeCadence', float),
        }
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True)
        avg_stroke_distance = fitfile.Distance.from_meters(self._get_field(activity_summary, 'avgStrokeDistance', float))
        paddle = {
            'activity_id'               : activity_id,
            'strokes'                   : self._get_field(activity_summary, 'strokes', float),
            'avg_stroke_distance'       : avg_stroke_distance.meters_or_feet(self.measurement_system),
        }
        PaddleActivities.s_insert_or_update(self.garmin_act_db_session, paddle, ignore_none=True)

    def _process_cycling(self, sub_sport, activity_id, activity_summary):
        activity = {
            'activity_id'               : activity_id,
            'avg_cadence'               : self._get_field(activity_summary, 'averageBikingCadenceInRevPerMinute', float),
            'max_cadence'               : self._get_field(activity_summary, 'maxBikingCadenceInRevPerMinute', float),
        }
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True)
        ride = {
            'activity_id'               : activity_id,
            'strokes'                   : self._get_field(activity_summary, 'strokes', float),
            'vo2_max'                   : self._get_field(activity_summary, 'vO2MaxValue', float),
        }
        CycleActivities.s_insert_or_update(self.garmin_act_db_session, ride, ignore_none=True)

    def _process_mountain_biking(self, sub_sport, activity_id, activity_summary):
        return self._process_cycling(sub_sport, activity_id, activity_summary)

    def _process_fitness_equipment(self, sub_sport, activity_id, activity_summary):
        root_logger.debug("process_fitness_equipment (%s) for %s", sub_sport, activity_id)
        self._call_process_func(sub_sport.name, None, activity_id, activity_summary)

    def _activities_process_json(self, json_data):
        activity_id = json_data['activityId']
        event = Event.from_json(json_data)
        sport, sub_sport = get_summary_sport(json_data)
        activity = {
            'activity_id'               : activity_id,
            'name'                      : json_data.get('activityName'),
            'description'               : self._get_field(json_data, 'description'),
            'type'                      : event.name,
            'sport'                     : sport.name,
            'sub_sport'                 : sub_sport.name,
            'laps'                      : self._get_field(json_data, 'lapCount'),
        }
        activity.update(self._process_common(json_data))
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True)
        self._call_process_func(sport.name, sub_sport, activity_id, json_data)
        return 1


class GarminJsonDetailsData(GarminJsonActivityData):
    """Class for importing Garmin activity data from JSON formatted Garmin Connect details downloads."""

    def __init__(self, db_params, input_dir, latest, measurement_system, debug):
        """
        Return an instance of GarminJsonDetailsData.

        Parameters:
        ----------
        db_params (dict): configuration data for accessing the database
        input_dir (string): directory (full path) to check for data files
        latest (Boolean): check for latest files only
        measurement_system (enum): which measurement system to use when importing the files
        debug (Boolean): enable debug logging

        """
        logger.info("Processing activities detail data")
        super().__init__(db_params, r'activity_details_\d*\.json', input_dir, latest, measurement_system, debug)

    def _process_steps_activity(self, sub_sport, activity_id, json_data):
        summary_dto = json_data['summaryDTO']
        avg_speed = fitfile.conversions.mps_to_mph(summary_dto.get('averageSpeed'))
        avg_moving_speed = fitfile.conversions.mps_to_mph(summary_dto.get('averageMovingSpeed'))
        max_speed = fitfile.conversions.mps_to_mph(summary_dto.get('maxSpeed'))
        run = {
            'activity_id'       : activity_id,
            'avg_pace'          : fitfile.conversions.perhour_speed_to_pace(avg_speed),
            'avg_moving_pace'   : fitfile.conversions.perhour_speed_to_pace(avg_moving_speed),
            'max_pace'          : fitfile.conversions.perhour_speed_to_pace(max_speed),
        }
        root_logger.debug("steps_activity for %d: %r", activity_id, run)
        StepsActivities.s_insert_or_update(self.garmin_act_db_session, run, ignore_none=True)

    def _process_cycling(self, sub_sport, activity_id, json_data):
        root_logger.debug("cycling (%s) for %d: %r", sub_sport, activity_id, json_data)

    def _process_elliptical(self, sub_sport, activity_id, json_data):
        root_logger.debug("elliptical for %d: %r", activity_id, json_data)

    def _process_hiking(self, sub_sport, activity_id, json_data):
        root_logger.debug("hiking for %d: %r", activity_id, json_data)
        self._process_steps_activity(sub_sport, activity_id, json_data)

    def _process_inline_skating(self, sub_sport, activity_id, json_data):
        root_logger.debug("inline_skating for %d: %r", activity_id, json_data)

    def _process_paddling(self, sub_sport, activity_id, json_data):
        root_logger.debug("paddling for %d: %r", activity_id, json_data)
    #
    # def _process_mountain_biking(self, sub_sport, activity_id, json_data):
    #     root_logger.debug("mountain_biking for %d: %r", activity_id, json_data)

    def _process_resort_skiing_snowboarding(self, sub_sport, activity_id, json_data):
        root_logger.debug("resort_skiing_snowboarding for %d: %r", activity_id, json_data)

    def _process_snowshoeing(self, sub_sport, activity_id, json_data):
        root_logger.debug("snow_shoe for %d: %r", activity_id, json_data)

    def _process_strength_training(self, sub_sport, activity_id, json_data):
        root_logger.debug("strength_training for %d: %r", activity_id, json_data)

    def _process_stand_up_paddleboarding(self, sub_sport, activity_id, json_data):
        root_logger.debug("stand_up_paddleboarding for %d: %r", activity_id, json_data)
    #
    # def _process_treadmill_running(self, sub_sport, activity_id, json_data):
    #     root_logger.debug("treadmill_running for %d: %r", activity_id, json_data)
    #     self._process_steps_activity(sub_sport, activity_id, json_data)

    def _process_running(self, sub_sport, activity_id, json_data):
        root_logger.debug("running (%s) for %d: %r", sub_sport, activity_id, json_data)
        self._process_steps_activity(sub_sport, activity_id, json_data)

    def _process_walking(self, sub_sport, activity_id, json_data):
        root_logger.debug("walking (%s) for %d: %r", sub_sport, activity_id, json_data)
        self._process_steps_activity(sub_sport, activity_id, json_data)

    def _process_fitness_equipment(self, sub_sport, activity_id, json_data):
        root_logger.debug("fitness_equipment (%s) for %d: %r", sub_sport, activity_id, json_data)
        self._call_process_func(sub_sport.name, None, activity_id, json_data)

    @classmethod
    def get_self_eval_feel(cls, value):
        """Return the Garmin Connect self evaluation 'How did you feel' label for the activity."""
        levels = [(100, "Very Strong"), (75, "Strong"), (50, "Normal"), (25, "Weak"), (0, "Very Weak")]
        for threshold, label in levels:
            if value >= threshold:
                return label

    @classmethod
    def get_self_eval_effort(cls, value):
        """Return the Garmin Connect self evaluation perceived effort label for the activity."""
        levels = [(100, "Maximum"), (90, "Extremely Hard"), (70, "Very Hard"), (50, "Hard"),
                  (40, "Somewhat Hard"), (30, "Moderate"), (20, "Light"), (10, "Very Light"), (0, "None")]
        for threshold, label in levels:
            if value >= threshold:
                return label

    def _activities_process_json(self, json_data):
        activity_id = json_data['activityId']
        metadata_dto = json_data['metadataDTO']
        summary_dto = json_data['summaryDTO']
        sport, sub_sport = get_details_sport(json_data)
        self_eval_feel = self._get_field(summary_dto, 'directWorkoutFeel', int)
        self_eval_effort = self._get_field(summary_dto, 'directWorkoutRpe', int)
        activity = {
            'activity_id'           : activity_id,
            'course_id'             : self._get_field(metadata_dto, 'associatedCourseId', int),
            'device_serial_number'  : self._get_field(metadata_dto, 'deviceId', int),
            'self_eval_feel'        : self.get_self_eval_feel(self_eval_feel) if (self_eval_feel is not None) else None,
            'self_eval_effort'      : self.get_self_eval_effort(self_eval_effort) if (self_eval_effort is not None) else None,
            'training_load'         : self._get_field(summary_dto, 'activityTrainingLoad', float)
        }
        activity.update(self._process_common(summary_dto))
        Activities.s_insert_or_update(self.garmin_act_db_session, activity, ignore_none=True)
        self._call_process_func(sport.name, sub_sport, activity_id, json_data)
        return 1



#!/usr/bin/env python3

"""Class running a checkup against the DB data."""

__author__ = "Tom Goetz"
__copyright__ = "Copyright Tom Goetz"
__license__ = "GPL"

import sys
import logging
from datetime import datetime, time, timedelta

import fitfile

from garmindb.garmindb import GarminDb, Attributes, Device, DeviceInfo, DailySummary, ActivitiesDb, Activities, StepsActivities


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler(stream=sys.stdout))
root_logger = logging.getLogger()


class Checkup():
    """Class running a checkup against the DB data."""

    def __init__(self, gc_config, paragraph_func=logger.info, heading_func=logger.info, debug=False):
        """Return an instance of the CheckUp class."""
        self.gc_config = gc_config
        self.db_params = self.gc_config.get_db_params()
        self.paragraph_func = paragraph_func
        self.heading_func = heading_func
        self.debug = debug
        self.garmin_db = GarminDb(self.db_params)
        self.measurement_system = Attributes.measurements_type(self.garmin_db)
        self.unit_strings = fitfile.units.unit_strings[self.measurement_system]

    def goals(self):
        """Do a checkup of the user's goals."""
        look_back_days = self.gc_config.get_node_value_default('checkup', 'look_back_days', 90)
        end_ts = datetime.now()
        start_ts = end_ts - timedelta(days=look_back_days)
        results = DailySummary.get_for_period(self.garmin_db, start_ts, end_ts)
        step_goal_days = 0
        step_goal_days_in_week = 0
        floors_goal_days = 0
        floor_goal_days_in_week = 0
        days_in_week = 0
        intensity_time = time.min
        intensity_time_goal = time.min
        intensity_weeks = 0
        intensity_goal_weeks = 0
        for result in results:
            if result.day.weekday() == 0:
                days_in_week = 0
                step_goal_days_in_week = 0
                floor_goal_days_in_week = 0
                intensity_time = time.min
                intensity_time_goal = time.min
            days_in_week += 1
            if result.steps_goal_percent >= 100:
                step_goal_days += 1
                step_goal_days_in_week += 1
            else:
                self.paragraph_func(f'Steps: goal not met on {result.day}')
            if result.floors_goal_percent >= 100:
                floors_goal_days += 1
                floor_goal_days_in_week += 1
            else:
                self.paragraph_func(f'Floors: goal not met on {result.day}')
            intensity_time = fitfile.conversions.add_time(intensity_time, result.intensity_time)
            intensity_time_goal = fitfile.conversions.add_time(intensity_time_goal, result.intensity_time_goal)
            if result.day.weekday() == 6:
                if days_in_week == 7:
                    intensity_weeks += 1
                    if step_goal_days_in_week < days_in_week:
                        self.paragraph_func(f'Steps: goal not met {days_in_week - step_goal_days_in_week} days for week ending in {result.day}')
                    if floor_goal_days_in_week < days_in_week:
                        self.paragraph_func(f'Floors: goal not met {days_in_week - floor_goal_days_in_week} days for week ending in {result.day}')
                    if intensity_time >= intensity_time_goal:
                        intensity_goal_weeks += 1
                    else:
                        self.paragraph_func(f'Intensity mins: goal not met for week ending in {result.day}')
        self.heading_func('Summary:')
        self.paragraph_func(f'Steps: met goal {step_goal_days} of last {look_back_days} days')
        self.paragraph_func(f'Floors: met goal {floors_goal_days} of last {look_back_days} days')
        self.paragraph_func(f'Intensity mins: met goal {intensity_goal_weeks} of last {intensity_weeks} weeks')

    def __activity_string(self, activity_db, activity):
        if activity.is_steps_activity():
            steps_activity = StepsActivities.get(activity_db, activity.activity_id)
            return ('%s: "%s" %.2f %s in %s pace: %s %s speed: %.2f %s' %
                    (activity.start_time, activity.name, activity.distance, self.unit_strings[fitfile.units.UnitTypes.distance_long], activity.elapsed_time,
                     steps_activity.avg_pace, self.unit_strings[fitfile.units.UnitTypes.pace], activity.avg_speed,
                     self.unit_strings[fitfile.units.UnitTypes.speed]))
        return '%s: "%s" %s in %s (%s)' % (activity.start_time, activity.name, activity.distance, activity.elapsed_time, activity.avg_speed)

    def activity_course(self, course_id):
        """Run a checkup on all activities matching the course_id."""
        activity_db = ActivitiesDb(self.db_params, self.debug)
        activities = Activities.get_by_course_id(activity_db, course_id)
        activities_count = len(activities)
        fastest_activity = Activities.get_fastest_by_course_id(activity_db, course_id)
        slowest_activity = Activities.get_slowest_by_course_id(activity_db, course_id)
        self.paragraph_func(f'Matching Activities: {activities_count}')
        self.paragraph_func(f'  first: {self.__activity_string(activity_db, activities[0])}')
        self.paragraph_func(f'  lastest: {self.__activity_string(activity_db, activities[-1])}')
        self.paragraph_func(f'  fastest: {self.__activity_string(activity_db, fastest_activity)}')
        self.paragraph_func(f'  slowest: {self.__activity_string(activity_db, slowest_activity)}')

    def battery_status(self):
        """Check for devices with low battery status."""
        devices = Device.get_all(self.garmin_db)
        for device in devices:
            battery_level = DeviceInfo.get_col_latest_where(self.garmin_db, DeviceInfo.battery_status,
                                                            [DeviceInfo.serial_number == device.serial_number,
                                                             DeviceInfo.battery_status != fitfile.field_enums.BatteryStatus.invalid])
            if battery_level is fitfile.field_enums.BatteryStatus.low:
                self.paragraph_func(f"Device {device.manufacturer} {device.product} ({device.serial_number}) has a low battery")
